/*
 * sortdemo.c
 *
 * version 0.1
 *
 * Anuradha Ratnaweera
 *
 * Dept. of Computer Sciences
 * Faculty of Engineering
 * University of Peradeniya
 * Peradeniya 20400
 * Sri Lanka
 *
 * anuradha@cs.pdn.ac.lk
 *
 * http://www.lklug.pdn.ac.lk/software/sortdemo/
 *
 * 05 April 2000
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <strings.h>
#include <gd.h>

#include "sortdemo.h"

void writeimage(int *, int, int, int, int, int, int);

int gifcount;

int main(int argc, char *argv[])
{
  void syntax();
  void initlist(int*, int, int);
  void qs(int*, int);
  void bubble1sort(int*, int);
  void bubble2sort(int*, int);
  void bubble3sort(int*, int);
  int writetofile(char*, int*, int);

  int a[N];
  int sorttype;
  int datatype;
  int n;

  /* check for 3 command line arguments */

  if (argc < 4){
    syntax();
    return 1;
  }

  /* set sort type from the first argument */
  
  if (strncmp(argv[1], "q", 1) == 0)
    sorttype = QUICKSORT;
  else if (strncmp(argv[1], "b1", 2) == 0)
    sorttype = BUBBLE1;
  else if (strncmp(argv[1], "b2", 2) == 0)
    sorttype = BUBBLE2;
  else if (strncmp(argv[1], "b3", 2) == 0)
    sorttype = BUBBLE3;
  else{
    fprintf(stderr, "\nIllegal sort type\n");
    syntax();
    return 2;
  }

  /* set data type from the second argument */

  if (strncmp(argv[2], "sor", 3) == 0)
    datatype = SORTED;
  else if (strncmp(argv[2], "rev", 3) == 0)
    datatype = REVERSE;
  else if (strncmp(argv[2], "ran", 3) == 0)
    datatype = RANDOM;
  else{
    fprintf(stderr, "\nIllegal data type\n");
    syntax();
    return 2;
  }

  /* set the size of the array from the third argument */

  n = atoi(argv[3]);
  if (n <= 0){
    fprintf(stderr, "\nThe array size should be greater than zero\n\n");
    return 3;
  }
  else if (n > N){
    fprintf(stderr, "\nThe array size should not be greater than %d\n\n", N);
    return 3;
  }

  /* initialize the list */

  initlist(a, datatype, n);

  /* write the unsorted list to a file for verifying the sort */

  writetofile(UNSORTEDFILE, a, n);

  /* initialize the gif settings */

  gifcount = 0;

  /* call the sorting routine */

  switch (sorttype){
    case QUICKSORT:
      qs(a, n);
      break;
    case BUBBLE1:
      bubble1sort(a, n);
      break;
    case BUBBLE2:
      bubble2sort(a, n);
      break;
    case BUBBLE3:
      bubble3sort(a, n);
      break;
  }

  /* write the sorted list to a file for verifying the sort */

  writetofile(SORTEDFILE, a, n);

  /* destroy the gif pointer */

  
  return 0;
}

void syntax()
/* prints the syntax of the program */
{
  fprintf(stderr, "\n");
  fprintf(stderr, "usage: %s type data size\n", PROGNAME);
  fprintf(stderr, "\n");
  fprintf(stderr, "type:\n");
  fprintf(stderr, "  q   quick sort\n");
  fprintf(stderr, "  b1  bubble sort\n");
  fprintf(stderr, "  b2  bubble sort with a flag\n");
  fprintf(stderr, "  b3  two way bubble sort with a flag\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "data: sorted, reverse or random\n");
  fprintf(stderr, "size: size of the array\n");
  fprintf(stderr, "\n");
}

void initlist(int *a, int type, int n)
/* initializes the list */
{
  int i;

  switch (type) {
    case SORTED:
      for (i = 0; i < n; i++)
	a[i] = i + DATAMIN;
      break;
    case REVERSE:
      for (i = 0; i < n; i++)
	a[i] = n - i + DATAMIN;
      break;
    case RANDOM:
      srand((unsigned int)time(NULL));
      for (i = 0; i < n; i++)
	a[i] = rand() % (DATAMAX - DATAMIN) + DATAMIN;
      break;
  }
}

void writeimage(int *a, int n, int n1, int n2, int n3, int n4, int testing)
{
  FILE *imagefile;
  char fname[100];
  int i, ii, j;
  gdImagePtr gifptr;
  int bgcolor, color, testcolor, swapcolor, listcolor, col;

  gifptr = gdImageCreate(n * BARSTEP + 1, IMAGEHEIGHT);
  bgcolor = gdImageColorAllocate(gifptr, 210, 210, 255);
  color = gdImageColorAllocate(gifptr, 0, 0, 0);
  testcolor = gdImageColorAllocate(gifptr, 50, 150, 50);
  swapcolor = gdImageColorAllocate(gifptr, 200, 50, 50);
  listcolor = gdImageColorAllocate(gifptr, 160, 160, 160);

  for (i = 0; i < n; i++){
    ii = i * BARSTEP;

    if ((i == n1) || (i == n2)){
      if (testing % 2){
	col = testcolor;
      }
      else{
	col = swapcolor;
      }
    }
    else{
      if ((i >= n3) && (i <= n4)){
	col = color;
      }
      else{
        col = listcolor;
      }
    }

    for (j = 1; j < BARSTEP; j++){
      gdImageLine(gifptr, ii+j, IMAGEHEIGHT - MARGIN - a[i],
		          ii+j, IMAGEHEIGHT - MARGIN, col);
    }
  }

  sprintf(fname, "sort%05d.gif", gifcount);

  imagefile = fopen(fname, "wb");
  gdImageGif(gifptr, imagefile);
  fclose(imagefile);

  gdImageDestroy(gifptr);
  
  gifcount++;
}

void qs(int a[], int n)
{
  void quicksort(int*, int, int, int);

  /* call the recursive sort routine on the whole array */

  quicksort(a, 0, n-1, n);

}

void quicksort(int a[], int n1, int n2, int n)
{
  int i, last, tmp;
  
  if (n1 >= n2) return;

  SWAP(a[n1], a[(n1+n2)/2])
  last = n1;
  for (i = n1 + 1; i <= n2; i++){
    writeimage(a, n, i, last, n1, n2, 1);
    if (a[i] < a[last]){
      SWAP(a[last], a[i])
      writeimage(a, n, last, i, n1, n2, 0);
      if (i != last+1){
        SWAP(a[i], a[last+1])
        writeimage(a, n, i, last+1, n1, n2, 0);
      }
      last++;
    }
  }

  quicksort(a, n1, last - 1, n);
  quicksort(a, last + 1, n2, n);

}


void bubble1sort(int a[], int n)
/* bubble sort */
{
  int i, j, jj, tmp;

  for (i = n - 1; i >= 0; i--){
    for (j = 0, jj = 1; j < i; j++, jj++){
      writeimage(a, n, j, jj, 0, n, 1);
      if (a[j] > a[jj]){
	SWAP(a[j], a[jj])
        writeimage(a, n, j, jj, 0, n, 0);
      }
    }
  }
}

void bubble2sort(int a[], int n)
/* bubble sort with a flag */
{
  int i, j, jj, tmp, sorted = 0;

  for (i = n - 1; (i >= 0) && (sorted == 0); i--){
    sorted = 1;
    for (j = 0, jj = 1; j < i; j++, jj++){
      writeimage(a, n, j, jj, 0, n, 1);
      if (a[j] > a[jj]){
	sorted = 0;
	SWAP(a[j], a[jj])
        writeimage(a, n, j, jj, 0, n, 0);
      }
    }
  }
}

void bubble3sort(int a[], int n)
{
  int i1 = 0, i2 = n - 1, j, jj, tmp, sorted = 0, count = 1;

  while (sorted == 0) {
    sorted = 1;
    if (count % 2){
      for (j = i1, jj = i1 + 1; j < i2; j++, jj++){
        writeimage(a, n, j, jj, 0, n, 1);
        if (a[j] > a[jj]){
	  sorted = 0;
	  SWAP(a[j], a[jj])
          writeimage(a, n, j, jj, 0, n, 0);
	}
      }
      i2--;
    }
    else{
      for (j = i2 - 1, jj = i2; j >= i1; j--, jj--){
        writeimage(a, n, j, jj, 0, n, 1);
        if (a[j] > a[jj]){
	  sorted = 0;
	  SWAP(a[j], a[jj])
          writeimage(a, n, j, jj, 0, n, 0);
	}
      }
      i1++;
    }
    count++;
  }
}

int writetofile(char* fname, int *a, int n)
{
  int i;
  FILE *outfile;

  outfile = fopen(fname, "wt");
  for (i = 0; i < n; i++){
    fprintf(outfile, "%d\n", a[i]);
  }
  fclose(outfile);
  
  return 0;
}
